import psycopg2
import psycopg2.extras
import os
import json
from werkzeug.security import generate_password_hash
from config import Config

db_uri = Config.DATABASE_URI
conn = psycopg2.connect(db_uri)
cursor = conn.cursor()

print(f"Initializing database at: {db_uri}")

print("Creating tables (if they don't exist)...")
cursor.execute("""
CREATE TABLE IF NOT EXISTS users (
    id SERIAL PRIMARY KEY,
    username TEXT UNIQUE NOT NULL,
    password_hash TEXT NOT NULL,
    role TEXT NOT NULL CHECK(role IN ('admin', 'user')),
    rss_token TEXT -- Added for personalized RSS feeds
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS agents (
    id SERIAL PRIMARY KEY,
    name TEXT UNIQUE NOT NULL,
    tags TEXT NOT NULL, -- Storing tags as comma-separated string for simplicity
    search_depth INTEGER NOT NULL DEFAULT 10 CHECK(search_depth BETWEEN 5 AND 100), -- Increased max depth
    is_enabled BOOLEAN NOT NULL DEFAULT TRUE, -- Use BOOLEAN for PostgreSQL
    agent_type TEXT NOT NULL DEFAULT 'news' CHECK(agent_type IN ('news')) -- Added agent type, default 'news'
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS user_subscriptions (
    user_id INTEGER NOT NULL,
    agent_id INTEGER NOT NULL,
    PRIMARY KEY (user_id, agent_id),
    FOREIGN KEY (user_id) REFERENCES users(id) ON DELETE CASCADE,
    FOREIGN KEY (agent_id) REFERENCES agents(id) ON DELETE CASCADE
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS app_settings (
    id INTEGER PRIMARY KEY CHECK (id = 1), -- Ensure only one row exists
    openai_api_key TEXT DEFAULT NULL,
    google_api_key TEXT DEFAULT NULL,
    anthropic_api_key TEXT DEFAULT NULL,
    google_cse_id TEXT DEFAULT NULL,
    deep_dive_enabled BOOLEAN DEFAULT FALSE,
    deep_dive_search_depth INTEGER DEFAULT 25,
    pipeline_interval_minutes INTEGER DEFAULT 60, -- Default interval in minutes
    editor_llm_provider TEXT DEFAULT 'openai' CHECK(editor_llm_provider IN ('openai', 'anthropic')) -- Added for Editor LLM selection
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS system_prompts (
    agent_name TEXT PRIMARY KEY CHECK(agent_name IN ('Manager', 'Collector', 'Editor')),
    prompt TEXT NOT NULL
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS processed_urls (
    url TEXT PRIMARY KEY NOT NULL,
    processed_timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS final_articles (
    id SERIAL PRIMARY KEY,
    title TEXT NOT NULL,
    summary TEXT NOT NULL,
    source_url TEXT,
    tags TEXT NOT NULL, -- Comma-separated tags relevant to the article
    created_timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
    image_filename TEXT -- Added column for local image filename
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS agent_logs (
    id SERIAL PRIMARY KEY,
    timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
    agent_name TEXT NOT NULL, -- e.g., Manager, Collector, Editor, or specific task ID
    level TEXT NOT NULL CHECK(level IN ('INFO', 'WARNING', 'ERROR', 'DEBUG')), -- Log level
    message TEXT NOT NULL, -- The main log message
    details TEXT -- Optional extra details, like stack traces or data dumps
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS user_favorites (
    user_id INTEGER NOT NULL,
    article_id INTEGER NOT NULL,
    PRIMARY KEY (user_id, article_id),
    FOREIGN KEY (user_id) REFERENCES users(id) ON DELETE CASCADE,
    FOREIGN KEY (article_id) REFERENCES final_articles(id) ON DELETE CASCADE
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS user_deleted_articles (
    user_id INTEGER NOT NULL,
    article_id INTEGER NOT NULL,
    PRIMARY KEY (user_id, article_id),
    FOREIGN KEY (user_id) REFERENCES users(id) ON DELETE CASCADE,
    FOREIGN KEY (article_id) REFERENCES final_articles(id) ON DELETE CASCADE
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS user_source_blacklist (
    user_id INTEGER NOT NULL,
    domain_name TEXT NOT NULL, -- Store the extracted domain name (e.g., example.com)
    PRIMARY KEY (user_id, domain_name),
    FOREIGN KEY (user_id) REFERENCES users(id) ON DELETE CASCADE
);
""")

# --- New Tables for URL Management ---
cursor.execute("""
CREATE TABLE IF NOT EXISTS failed_urls (
    id SERIAL PRIMARY KEY,
    url TEXT NOT NULL UNIQUE,
    domain_name TEXT NOT NULL, -- Extracted domain for easier grouping/blacklisting
    error_type TEXT NOT NULL, -- e.g., '403 Forbidden', 'Timeout', 'Scrape Error', 'Validation Error'
    failure_count INTEGER DEFAULT 1,
    last_attempt_timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
);
""")
cursor.execute("""
CREATE TABLE IF NOT EXISTS system_domain_blacklist (
    domain_name TEXT PRIMARY KEY NOT NULL, -- e.g., example.com
    reason TEXT, -- Optional: Why it was blacklisted
    added_timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP
);
""")

# --- New Table for Agent Run Times ---
cursor.execute("""
CREATE TABLE IF NOT EXISTS agent_run_logs (
    id SERIAL PRIMARY KEY,
    agent_name TEXT NOT NULL, -- e.g., 'Pipeline', 'Collector', 'Editor', 'Manager:Tech News'
    start_timestamp TIMESTAMP WITH TIME ZONE NOT NULL,
    end_timestamp TIMESTAMP WITH TIME ZONE NOT NULL,
    duration_seconds REAL NOT NULL, -- Use REAL for floating-point seconds
    success BOOLEAN NOT NULL
);
""")

# Create api_usage_logs table
print("Creating api_usage_logs table...")
cursor.execute("""
CREATE TABLE IF NOT EXISTS api_usage_logs (
    id SERIAL PRIMARY KEY,
    timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
    api_provider TEXT NOT NULL, -- e.g., 'OpenAI', 'Google', 'Anthropic'
    service_model TEXT NOT NULL, -- e.g., 'gpt-4o', 'Custom Search', 'claude-3-haiku'
    caller_component TEXT, -- e.g., 'Collector', 'Editor', 'APITest'
    success BOOLEAN NOT NULL,
    input_tokens INTEGER, -- Nullable for non-LLM calls
    output_tokens INTEGER, -- Nullable for non-LLM calls
    metadata JSONB -- For any extra details, like error messages
);
""")

# Add default users
print("Adding default users...")
try:
    # Default Admin User
    admin_username = 'admin'
    admin_password = 'Superior14me!'
    admin_password_hash = generate_password_hash(admin_password)
    cursor.execute("INSERT INTO users (username, password_hash, role) VALUES (%s, %s, %s) ON CONFLICT (username) DO NOTHING",
                   (admin_username, admin_password_hash, 'admin'))
    print(f"Created admin user: {admin_username} / {admin_password}")

    # Default Regular User
    user_username = 'user'
    user_password = 'Superior14me!' # Use a stronger password in production!
    user_password_hash = generate_password_hash(user_password)
    cursor.execute("INSERT INTO users (username, password_hash, role) VALUES (%s, %s, %s) ON CONFLICT (username) DO NOTHING",
                   (user_username, user_password_hash, 'user'))
    print(f"Created regular user: {user_username} / {user_password}")

except psycopg2.IntegrityError as e:
    print(f"Warning: Could not insert default users (they might already exist). Error: {e}")

# Add default agents from JSON file
print("Adding default agents from default_agents.json...")
agents_file_path = os.path.join(os.path.dirname(__file__), 'default_agents.json')

try:
    with open(agents_file_path, 'r') as f:
        default_agents = json.load(f)

    for agent in default_agents:
        cursor.execute(
            "INSERT INTO agents (name, tags, search_depth, is_enabled, agent_type) VALUES (%s, %s, %s, %s, %s) ON CONFLICT (name) DO NOTHING",
            (agent['name'], agent['tags'], agent['search_depth'], agent['is_enabled'], agent['agent_type'])
        )
        print(f"Added/Skipped default agent: '{agent['name']}' (type: {agent['agent_type']})")

except psycopg2.IntegrityError as e:
     print(f"Warning: Could not insert default agents (they might already exist). Error: {e}")
except FileNotFoundError:
    print(f"Error: default_agents.json not found at {agents_file_path}. Skipping default agent creation.")
except json.JSONDecodeError:
    print(f"Error: Could not decode JSON from {agents_file_path}. Skipping default agent creation.")
except Exception as e:
    print(f"An unexpected error occurred while loading default agents: {e}")

# Add default settings row (with empty keys)
print("Adding default settings row...")
try:
    cursor.execute("INSERT INTO app_settings (id, openai_api_key, google_api_key, anthropic_api_key, google_cse_id, deep_dive_enabled, deep_dive_search_depth, pipeline_interval_minutes, editor_llm_provider) VALUES (1, NULL, NULL, NULL, NULL, FALSE, 25, 60, 'openai') ON CONFLICT (id) DO NOTHING")
    print("Added default settings row (including editor_llm_provider).")
except psycopg2.IntegrityError as e:
    print(f"Warning: Could not insert default settings (it might already exist). Error: {e}")

# Add default system prompts
print("Adding default system prompts...")
default_prompts = {
    'Manager': """You are the Manager agent. Your role is to orchestrate the news collection and editing process based on incoming articles related to specific tags.

1.  Receive a list of new potential articles (URL, title, content) from the Collector agent for a set of tags.
2.  **Noteworthiness Evaluation (If 'Deep Dive' Enabled in Settings):**
    *   For *each* incoming article, evaluate its potential significance and novelty. Consider factors like: Is this breaking news? A major event? Unexpected development? Highly impactful announcement? Does the title/summary suggest something beyond routine updates for the given tags?
    *   Respond for *each article* with a JSON object: `{{"noteworthy": true/false, "reason": "Brief justification"}}`. Be critical; only flag truly exceptional news as noteworthy.
3.  **Deep Dive Trigger (If 'Deep Dive' Enabled & Article is Noteworthy):**
    *   If an article is evaluated as `noteworthy: true`:
        a. Log that a deep dive is being triggered for this topic.
        b. Extract the core topic keywords or a concise search query from the noteworthy article's title and content. Respond with a JSON object: `{{"keywords": ["keyword1", "keyword2", "etc"]}}` or `{{"search_query": "concise query"}}`.
        c. Initiate a *second* Collector agent run using these keywords/query and the configured 'Deep Dive Search Depth'.
        d. Combine the original noteworthy article with all articles found in the deep dive.
        e. Send this combined list *directly* to the Editor agent for merging into a single comprehensive summary.
        f. Add the Editor's result (if successful) to the list of final articles to be saved. Mark all source URLs from this deep dive group as processed.
        g. *Skip* the standard similarity check (Step 4) for the original noteworthy article and any duplicates found during the deep dive.
4.  **Standard Processing (For Articles Not Triggering Deep Dive):**
    *   Group the remaining (non-deep-dived) articles based on title similarity (e.g., using a threshold like 0.7).
    *   If a group contains multiple similar articles: Delegate the group to the Editor agent for merging and summarizing.
    *   If an article has no significant similarity to others: Prepare it for saving directly (use its content, perhaps summarized or truncated if very long).
5.  **Finalization:**
    *   Collect all results (from deep dives and standard processing).
    *   Save each final article (title, summary, primary source URL, tags) to the `final_articles` database.
    *   Ensure all source URLs corresponding to saved articles (including all originals from merged groups/deep dives) are added to the `processed_urls` database.
    *   Log your actions comprehensively, including evaluations, deep dive triggers, similarity checks, and saving steps.

Your primary outputs during the process will be evaluation/keyword extraction JSONs when needed, and ultimately saving articles to the database. You coordinate calls to Collector and Editor.""",
    'Collector': """You are the Collector agent. Your role is to find relevant news articles based on given tags and extract their content.

1. Receive a list of news tags OR a specific search query.
2. Use the 'google_search' tool with num_results={search_depth} to find recent news articles matching these tags. Prioritize reputable sources. Formulate your search query by prepending 'latest news' or 'recent news articles' to the tags.
3. **Process Each Search Result:** For each result ('title', 'link', 'snippet'):
    a. **Validate URL/Title:** Analyze the 'link' and 'title'. If it looks like a homepage, category page, or generic site index (not a specific article), **SKIP THIS RESULT**. Log the skip.
    b. **Check System Blacklist:** Extract the domain from the 'link'. If the domain is blacklisted (check using domain name, e.g., 'example.com'), **SKIP THIS RESULT**. Log the skip.
    c. **Check Processed URL:** Use 'check_if_url_is_processed' for the 'link'. If processed, **SKIP THIS RESULT**. Log the skip.
    d. **Attempt RSS First:** If the URL is new and not blacklisted:
        i. Use 'find_rss_feed' on the base domain of the 'link'.
        ii. If a feed URL is found, use 'parse_rss_feed'. Check the entries in the feed. For each entry:
            - Check if the entry's 'link' (the actual article URL) has been processed using 'check_if_url_is_processed'.
            - If the entry's link is new, format it as a result object: `{{"url": entry_link, "title": entry_title, "content": entry_summary}}`. Add this object to your final list. **Do not scrape the page if you get content from RSS.** Mark the entry's link as processed in your internal tracking for this run.
    e. **Fallback to Scraping:** If no RSS feed was found for the domain OR if the specific article 'link' from the search result was NOT found in the RSS feed entries (after checking processed status), then use 'scrape_web_content' on the original 'link' from the search result.
        - If scraping is successful, format the result: `{{"url": scraped_url, "title": scraped_title, "content": scraped_content}}`. Add this object to your final list.
        - Handle scraping errors gracefully (log the error and skip).
4. **Final Output:** Return your findings as a single JSON list of objects. Each object represents ONE successfully processed NEW article (either from RSS or scraping) and MUST have the keys: "url", "title", and "content". If no new articles are found, return an empty list [].
5. Log your actions comprehensively, including searches, skips (validation, blacklist, processed), RSS attempts, and scraping attempts/successes/failures.""",
    'Editor': """You are the Editor agent. Your role is to rewrite and merge potentially overlapping or related news articles into a single, concise summary.
1. Receive a set of articles (each with URL, title, and full scraped content) identified by the Manager as overlapping or part of a deep dive on a single topic.
2. Analyze the articles to identify the core news event and key details. Synthesize the information from all provided articles.
3. **Actively remove** any extraneous information such as relative timestamps (e.g., '5 minutes ago', 'yesterday'), author names (unless crucial to the news), navigation links, ads, website boilerplate text, comment sections, or other non-essential text scraped from the source pages. Focus *only* on the core news information from the provided articles. Ensure the summary is neutral and factual.
4. **Rewrite and synthesize** the essential information into a single, well-written, concise, and smoothly flowing news summary (aim for 3-5 sentences unless the topic is exceptionally complex). The summary must be **original writing**, not just copied sentences with minor changes.
5. Create a suitable, informative title for the merged article that reflects the core event.
7. Return the result as a single JSON object with exactly two keys: "merged_title" and "merged_summary".
8. Log your actions."""
}

try:
    for name, prompt in default_prompts.items():
        # Use PostgreSQL's ON CONFLICT clause for "upsert" behavior
        cursor.execute("""
            INSERT INTO system_prompts (agent_name, prompt) VALUES (%s, %s)
            ON CONFLICT (agent_name) DO UPDATE SET prompt = EXCLUDED.prompt
        """, (name, prompt))
    print("Added/Updated default system prompts for Manager, Collector, and Editor.")
except psycopg2.IntegrityError as e:
    print(f"Warning: Could not insert/update default system prompts. Error: {e}")

# Subscribe admin to default agents
print("Subscribing admin user to default agents...")
try:
    # Get admin user ID
    cursor_dict = conn.cursor(cursor_factory=psycopg2.extras.DictCursor) # Use DictCursor for fetching
    cursor_dict.execute("SELECT id FROM users WHERE username = %s", ('admin',))
    admin_user_row = cursor_dict.fetchone()
    
    if not admin_user_row: 
        print("Warning: Could not find admin user ID. Skipping default subscriptions.")
    else:
        admin_user_id = admin_user_row[0]

        # Get agent IDs
        # Fetch IDs for the agents we want to subscribe the admin to by default
        # (Adjust names if they changed in the JSON file)
        default_agent_names_for_admin = ['Tech News Collector', 'Fashion Collector', 'Global Economics Collector', 'World Conflicts Collector']
        
        # Execute a single query to get all matching agents
        cursor_dict.execute("SELECT id, name FROM agents WHERE name = ANY(%s)", (default_agent_names_for_admin,))
        agent_rows = cursor_dict.fetchall()
        
        if not agent_rows:
            print("Warning: Could not find default agent IDs. Skipping default subscriptions.")
        else:
            for agent_row in agent_rows:
                agent_id = agent_row[0]
                agent_name = agent_row[1]
                # Insert subscription (IGNORE if it already exists)
                cursor.execute("INSERT INTO user_subscriptions (user_id, agent_id) VALUES (%s, %s) ON CONFLICT (user_id, agent_id) DO NOTHING", 
                              (admin_user_id, agent_id))
                print(f"Ensured admin user (ID: {admin_user_id}) is subscribed to {agent_name} (ID: {agent_id}).")
        
        cursor_dict.close() # Close the DictCursor

except psycopg2.Error as e:
    print(f"Warning: Error during default subscription setup: {e}")

# Create api_usage_logs table
print("Creating api_usage_logs table...")
cursor.execute("""
CREATE TABLE IF NOT EXISTS api_usage_logs (
    id SERIAL PRIMARY KEY,
    timestamp TIMESTAMP WITH TIME ZONE DEFAULT CURRENT_TIMESTAMP,
    api_provider TEXT NOT NULL, -- e.g., 'OpenAI', 'Google', 'Anthropic'
    service_model TEXT NOT NULL, -- e.g., 'gpt-4o', 'Custom Search', 'claude-3-haiku'
    caller_component TEXT, -- e.g., 'Collector', 'Editor', 'APITest'
    success BOOLEAN NOT NULL,
    input_tokens INTEGER, -- Nullable for non-LLM calls
    output_tokens INTEGER, -- Nullable for non-LLM calls
    metadata JSONB -- For any extra details, like error messages
);
""")

# Commit changes and close the connection
conn.commit()
conn.close()

print("Database initialization/update process completed.")
